// Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef OPEN_SPIEL_PUBLIC_STATES_GAME_TRANSFORMS_PERF_RECALL_FACT_OBSERVATIONS_H_
#define OPEN_SPIEL_PUBLIC_STATES_GAME_TRANSFORMS_PERF_RECALL_FACT_OBSERVATIONS_H_

// This is a game transform that takes Base API State that provides factored
// observations and produces perfect-recall Public State API representation.

namespace open_spiel {
namespace public_states {
namespace game_transforms {

// TODO(author13): This is just a stub.

}  // namespace game_transforms
}  // namespace public_states
}  // namespace open_spiel

#endif  // OPEN_SPIEL_PUBLIC_STATES_GAME_TRANSFORMS_PERF_RECALL_FACT_OBSERVATIONS_H_
